(function(){

	// Requirements
	var
		sys = require("sys"),
		fs = require("fs"),
		path = require("path"),
		jshint = false,
		less = false,
		pulverizr = false,
		uglify = false,
			jsp = false,
			pro = false,
		util = require('util'),
		exec = require('child_process').exec,
		deps = require(__dirname+'/dep.js'),
		util = require(__dirname+'/util.js');

	// Buildr Constructor
	var Buildr = function(){
		this.config = {}.extend(this.config);
		this.init.apply(this,arguments);
	};

	// Buildr Definition
	Buildr.prototype.extend({
		/**
		 * Configuration
		 */
		"config": {
			// User Configurable
			"compress": {
				"js": true,
				"css": true,
				"img": false,
				"html": true
			},
			"check": {
				"js": false,
				"css": false,
				"jsOptions": {
					"browser": true
				}
			},
			"bundle": {
				"js": true,
				"css": true,
				"src": false
			},
			"directories": {
				"out": "./out",
				"src": "./src",
				"templates": "./templates",
				"subpackages": "subpackages"
			},
			"templates": {
				"out_bundle_header.js": "out_bundle_header.js",
				"out_bundle_footer.js": "out_bundle_footer.js",
				"out_bundle_item.js": "out_bundle_item.js",
				"out_bundle_subpackage.js": "out_bundle_subpackage.js",
				"src_bundle_header.js": "src_bundle_header.js",
				"src_bundle_footer.js": "src_bundle_footer.js",
				"src_bundle_item.js": "src_bundle_item.js",
				"src_bundle_subpackage.js": "src_bundle_subpackage.js",

				"out_bundle_header.css": "out_bundle_header.css",
				"out_bundle_footer.css": "out_bundle_footer.css",
				"out_bundle_item.css": "out_bundle_item.css",
				"out_bundle_subpackage.css": "out_bundle_subpackage.css",
				"src_bundle_header.css": "src_bundle_header.css",
				"src_bundle_footer.css": "src_bundle_footer.css",
				"src_bundle_item.css": "src_bundle_item.css",
				"src_bundle_subpackage.css": "src_bundle_subpackage.css"
			},
			"subpackages": true,
			"ignore": {
				"js": [],
				"css": [],
				"img": [],
				"check": []
			},
			"files": {
				"js": true,
				"css": true,
				"img": true
			},

			// Autogenerated
			"templateContent": {
				"out_bundle_item.js": "%DATA%",
				"out_bundle_item.css": "%DATA%",
				"src_bundle_item.js": "",
				"src_bundle_item.css": "@import url(%URL%);\n",
			}
		},

		/**
		 * Initialise the Builder
		 * @param {String|Object} options
		 */
		init: function(options){
			// Configure
			if ( typeof options === 'string' ) {
				options = this.loadOptions(options);
			}
			this.applyOptions(options);
		},

		// Define
		expandPath: function(value,rootPath,relativePath){
			// Full Path
			if ( value.beginsWith('/') ) {
				// do nothing
			}
			// Relative to Root Directory
			else if ( value.beginsWith('./') ) {
				value = value.replace(/^\.\//, rootPath+'/');
			}
			// Relative to Relative Path
			else {
				value = relativePath+'/'+value;
			}

			// Apply new value
			return value;
		},

		// Define
		shrinkPath: function(value){
			var value = util.getRelativePath(value,[this.config.directories.src,this.config.directories.out]);
			return value;
		},
		isIgnored: function(ignores,value){
			var skip = false;
			ignores.each(function(i,ignore){
				if ( value.beginsWith(ignore) ) {
					skip = true;
					return false;
				}
			});
			return skip;
		},

		/**
		 * Load the Options
		 * @param {String} rootPath
		 * @return {Object} options
		 */
		loadOptions: function(rootPath){
			// Prepare
			var options = {};

			// Path
			options.rootPath = fs.realpathSync(rootPath.strip('package.json'));
			options.metaPath = options.rootPath+'/package.json';

			// Check if the meta data exists
			var exists = util.fileExistsSync(options.metaPath);
			if ( exists ) {
				// There is already meta data
				options.extend(JSON.parse(fs.readFileSync(options.metaPath).toString()).buildr||{});
			}

			// Return
			return options;
		},

		/**
		 * Apply the Options
		 * @param {Object} options
		 */
		applyOptions: function(options){
			// Prepare
			var config = this.config, me = this;

			// Check
			if ( typeof options !== 'object' ) {
				abort('Options was not of a valid type');
			}

			// Apply Options to Configuration
			config.extend(options);

			// ----------------------------------------------------------------------
			// Handle Shortcuts

			// Compress
			if ( config.compress === true || config.compress === false ) {
				config.compress = {
					"js": config.compress,
					"css": config.compress,
					"img": config.compress,
					"html": config.compress
				};
			}

			// Bundle
			if ( config.bundle === true || config.bundle === false ) {
				config.bundle = {
					"js": config.bundle,
					"css": config.bundle
				};
			}

			// Name
			config.name = config.name || path.basename(config.rootPath);

			// ----------------------------------------------------------------------
			// Expand Directory Paths

			// Bundle Javascript
			if ( config.bundle.js === true ) {
				config.bundle.js = config.name+'.js';
			}

			// Bundle CSS
			if ( config.bundle.css === true ) {
				config.bundle.css = config.name+'.css';
			}

			// Directories
			config.directories.each(function(key,value){
				config.directories[key] = me.expandPath(value,config.rootPath,config.directories.src);
			});

			// Templates
			config.templates.each(function(key,value){
				config.templates[key] = me.expandPath(value,config.rootPath,config.directories.templates);
			});

			// Subpackages
			config.subpackages.each(function(key,value){
				config.subpackages[key] = me.expandPath(value,config.rootPath,config.directories.subpackages);
			});

			// ----------------------------------------------------------------------
			// Detect Files

			// JS
			if ( config.files.js === true ) {
				// Scan the directory for javascript files
				config.files.js = util.findFilesWithExtensionSync(config.directories.src, ".js");
			}

			// CSS
			if ( config.files.css === true ) {
				// Scan the directory for css files
				config.files.css = util.findFilesWithExtensionSync(config.directories.src, ".css");
			}

			// IMG
			if ( config.files.img === true ) {
				// Scan the directory for image files
				config.files.img = util.findFilesWithExtensionSync(config.directories.src, (/\.(jpe?g|png|w?bmp|[gt]iff?)$/i));
			}

			// ----------------------------------------------------------------------
			// Merge in Subpackages

			// Cycle
			config.subpackages.each(function(key,subpackagePath){
				// Prepare
				var subBuilder = new Buildr(subpackagePath),
					jsFiles = subBuilder.config.files.js,
					cssFiles = subBuilder.config.files.css,
					imgFiles = subBuilder.config.files.img;

				// Adjust Paths
				jsFiles.each(function(i,filePath){
					jsFiles[i] = me.expandPath(filePath,subBuilder.config.rootPath,subBuilder.config.directories.src);
				});
				cssFiles.each(function(i,filePath){
					cssFiles[i] = me.expandPath(filePath,subBuilder.config.rootPath,subBuilder.config.directories.src);
				});
				imgFiles.each(function(i,filePath){
					imgFiles[i] = me.expandPath(filePath,subBuilder.config.rootPath,subBuilder.config.directories.src);
				});

				// Attach Files
				config.files.js.extend(jsFiles);
				config.files.css.extend(cssFiles);
				config.files.img.extend(imgFiles);
			});

			// ----------------------------------------------------------------------
			// Shrink File Paths

			// JS
			config.files.js.each(function(key,value){
				value = me.shrinkPath(value);
				if ( me.isIgnored(config.ignore.js, value) ) {
					value = false;
				}
				config.files.js[key] = value;
			});

			// CSS
			config.files.css.each(function(key,value){
				value = me.shrinkPath(value);
				if ( me.isIgnored(config.ignore.css, value) ) {
					value = false;
				}
				config.files.css[key] = value;
			});

			// IMG
			config.files.img.each(function(key,value){
				value = me.shrinkPath(value);
				if ( me.isIgnored(config.ignore.img, value) ) {
					value = false;
				}
				config.files.img[key] = value;
			});

			// ----------------------------------------------------------------------
			// Fetch Template Content

			// Templates
			config.templates.each(function(key,value){
				var templateData = (value && util.fileExistsSync(value)) ? fs.readFileSync(value).toString() : '';
				config.templateContent[key] = templateData || config.templateContent[key] || '';
			});


		},

		/**
		 * Compile our Project
		 */
		compile: function(){
			// Prepare
			var config = this.config;

			// Check
			this.check();

			// Clean
			if ( config.directories.out ) {
				console.log("Cleaning...");
				if ( util.fileExistsSync(config.directories.out) ) {
					util.rmdirSync(config.directories.out);
				}
				fs.mkdirSync(config.directories.out,0700);
				util.cpdirSync(config.directories.src,config.directories.out);
			}

			// Bundle
			this.bundle();

			// Compress
			this.compress();

			// Compare
			var
				srcStats = util.dirstatsSync(config.directories.src),
				outStats = util.dirstatsSync(config.directories.out),
				percent = Math.round((outStats.size/srcStats.size)*100);

			// Output Compare
			console.log("\nReduced source from\n\t["+srcStats.files+"] files and ["+srcStats.size+"] bytes to\n\t["+outStats.files+"] files and ["+outStats.size+"] bytes\n\t["+percent+"%] of original size");

			// Log
			console.log("\nCompilation Complete :)");
		},

		/**
		 * Check all JS Files
		 */
		checkJs: function(){
			// Prepare
			var
				config = this.config, me = this,
				foundAll = false;

			// Load
			jshint = require("jshint").JSHINT;

			// Check
			config.files.js.each(function(i,filePath){
				// Ignore?
				if ( !filePath || me.isIgnored(config.ignore.check, filePath) ) { return; }

				// Prepare
				var
					fileSrcPath = me.expandPath(filePath,config.rootPath,config.directories.src),
					fileData = fs.readFileSync(fileSrcPath).toString(),
					result, foundOne = false;

				// Output
				jshint(fileData,config.check.jsOptions);
				result = jshint.data();

				// Result
				if ( !result.errors || !result.errors.length ) { return; }

				console.log(filePath);

				// Output errors
				result.errors.each(function(i,error){
					if ( !error || !error.raw ) { return; }
					foundOne = true;
					console.log(
							'\tLine '+error.line+':'+
							' '+error.raw.replace(/\{([a-z])\}/,function(a,b){
								return error[b]||a;
							})+
							(error.evidence ? '\n\t'+error.evidence.replace(/^\s+/,'') : '')+
							'\n'
					);
				});

				// Fail
				if ( foundOne ) {
					process.exit();
				}
			});

			// Fail
			if ( foundAll ) {
				process.exit();
			}
		},

		/**
		 * Check the Files
		 */
		check: function() {
			// Prepare
			var config = this.config;

			// JS
			if ( config.check.js ) {
				console.log("\nChecking JS Files:");
				this.checkJs();
			}

			// CSS
			if ( config.bundle.css ) {
				console.log("\nChecking CSS Files not implemented yet.");
				//console.log("\nChecking CSS Files:");
				//this.checkCss();
			}
		},

		/**
		 * Bundle all JS Files
		 */
		bundleJs: function(){
			// Prepare
			var
				config = this.config, me = this,
				bundleSrcData = '', bundleSrcPath = config.directories.src+'/'+config.bundle.js,
				bundleOutData = '', bundleOutPath = config.directories.out+'/'+config.bundle.js;

			// Amend
			bundleSrcData += config.templateContent['src_bundle_header.js'];
			bundleOutData += config.templateContent['out_bundle_header.js'];

			// Subpackages
			config.subpackages.each(function(i,subpackage){
				if ( !subpackage ) return; // continue
				console.log("Adding: "+path.basename(subpackage));

				// Amend Src Text
				bundleSrcData += config.templateContent['src_bundle_subpackage.js']
					.replace('%NAME%',path.basename(subpackage));
				bundleOutData += config.templateContent['out_bundle_subpackage.js']
					.replace('%NAME%',path.basename(subpackage));
			});

			// Bundle
			config.files.js.each(function(i,filePath){
				console.log("Bundled: "+filePath);

				// Prepare
				var
					fileOutPath = me.expandPath(filePath,config.rootPath,config.directories.out),
					fileData = fs.readFileSync(fileOutPath).toString();

				// Amend Files
				bundleOutData += config.templateContent['out_bundle_item.js']
					.replace('%DATA%',fileData)
					.replace('%URL%',filePath);
				bundleSrcData += config.templateContent['src_bundle_item.js']
					.replace('%DATA%',fileData)
					.replace('%URL%',filePath);

				// Delete the Out File
				fs.unlinkSync(fileOutPath);
			});

			// Amend
			bundleSrcData += config.templateContent['src_bundle_footer.js'];
			bundleOutData += config.templateContent['out_bundle_footer.js'];

			// Write Files
			console.log("\nBundling JS Files to: "+bundleOutPath);
			fs.writeFileSync(bundleOutPath,bundleOutData);
			if ( config.bundle.src ) {
				console.log("Bundling JS Files to: "+bundleSrcPath);
				fs.writeFileSync(bundleSrcPath,bundleSrcData);
			}

			// Add Files
			config.files.js = [];
			config.files.js.push(me.shrinkPath(bundleOutPath));
		},

		/**
		 * Bundle all Css Files
		 */
		bundleCss: function(){
			// Prepare
			var
				config = this.config, me = this,
				bundleSrcData = '', bundleSrcPath = config.directories.src+'/'+config.bundle.css,
				bundleOutData = '', bundleOutPath = config.directories.out+'/'+config.bundle.css;

			// Amend
			bundleSrcData += config.templateContent['src_bundle_header.css'];
			bundleOutData += config.templateContent['out_bundle_header.css'];

			// Subpackages
			config.subpackages.each(function(i,subpackage){
				if ( !subpackage ) return; // continue
				console.log("Adding: "+path.basename(subpackage));

				// Amend Src Text
				bundleSrcData += config.templateContent['src_bundle_subpackage.css']
					.replace('%NAME%',path.basename(subpackage));
				bundleOutData += config.templateContent['out_bundle_subpackage.css']
					.replace('%NAME%',path.basename(subpackage));
			});

			// Bundle
			config.files.css.each(function(i,filePath){
				if ( !filePath ) return; // continue
				console.log("Bundled: "+filePath);

				// Prepare
				var
					fileOutPath = me.expandPath(filePath,config.rootPath,config.directories.out),
					fileSrcPath = me.expandPath(filePath,config.rootPath,config.directories.src),
					fileData = fs.readFileSync(fileOutPath).toString(),
					parentSrcPath = fileSrcPath.replace(/[\/\\][^\/\\]+$/,'');

				// Amend Out Text
				fileData = fileData
					.replace(
						/url\(([^\)]+)\)/g,
						function(str, p1, offset){
							// Trim quotes
							var url = p1.replace(/[\'\"]/g,'');

							// Update url
							if ( url[0] !== '/' && !/\:/.test(url) ) {
								url = fs.realpathSync(parentSrcPath+'/'+url).replace(config.directories.src+'/','');
							}

							// Replace with new url
							return 'url(\''+url+'\')';
						}
					);

				// Amend Files
				bundleOutData += config.templateContent['out_bundle_item.css']
					.replace('%DATA%',fileData)
					.replace('%URL%',filePath);
				bundleSrcData += config.templateContent['src_bundle_item.css']
					.replace('%DATA%',fileData)
					.replace('%URL%',filePath);

				// Delete the Out File
				fs.unlinkSync(fileOutPath);
			});

			// Amend
			bundleSrcData += config.templateContent['src_bundle_footer.css'];
			bundleOutData += config.templateContent['out_bundle_footer.css'];

			// Write Files
			console.log("\nBundling CSS Files to: "+bundleOutPath);
			fs.writeFileSync(bundleOutPath,bundleOutData);
			if ( config.bundle.src ) {
				console.log("Bundling CSS Files to: "+bundleSrcPath);
				fs.writeFileSync(bundleSrcPath,bundleSrcData);
			}

			// Add Files
			config.files.css = [];
			config.files.css.push(me.shrinkPath(bundleOutPath));
		},

		/**
		 * Bundle the Files
		 */
		bundle: function() {
			// Prepare
			var config = this.config;

			// JS
			if ( config.bundle.js ) {
				console.log("\nBundling JS Files:");
				this.bundleJs();
			}

			// CSS
			if ( config.bundle.css ) {
				console.log("\nBundling CSS Files:");
				this.bundleCss();
			}
		},

		/**
		 * Compress JS Files
		 */
		compressJs: function(){
			// Prepare
			var config = this.config, me = this;

			// Load
			uglify = require("uglify-js");
				jsp = uglify.parser;
				pro = uglify.uglify;

			// Compress
			config.files.js.each(function(key,filePath){
				if ( !filePath ) return; // continue
				console.log("Compressed: "+filePath);
				var
					fileOutPath = me.expandPath(filePath,config.rootPath,config.directories.out),
					orig_code = fs.readFileSync(fileOutPath).toString(),
					ast = jsp.parse(orig_code); // parse code and get the initial AST
				ast = pro.ast_mangle(ast); // get a new AST with mangled names
				ast = pro.ast_squeeze(ast); // get an AST with compression optimizations
				var final_code = pro.gen_code(ast); // compressed code here
				fs.writeFileSync(fileOutPath,final_code);
			});
		},

		/**
		 * Compress CSS Files
		 */
		compressCss: function(){
			// Prepare
			var config = this.config, me = this;

			// Load
			less = require("less");

			// Compress
			config.files.css.each(function(key,filePath){
				if ( !filePath ) return; // continue
				console.log("Compressed: "+filePath);

				// Prepare
				var
					fileOutPath = me.expandPath(filePath,config.rootPath,config.directories.out),
					data = fs.readFileSync(fileOutPath).toString(),
					options = {
						paths: [config.directories.out],
						optimization: 1,
						filename: fileOutPath
					};

				// Minify
				new(less.Parser)(options).parse(data,
					function(err,tree){
						// Check
						if ( err ) {
							less.writeError(err, options);
							process.exit(1);
						} else {
							// Save
							try {
								css = tree.toCSS({
									compress: 1
								});
								fs.writeFileSync(fileOutPath,css);
							} catch (e) {
								less.writeError(e, options);
								process.exit(2);
							}
						}
					}
				);
			});
		},

		/**
		 * Compress IMG Files
		 */
		compressImg: function(){
			// Prepare
			var config = this.config, me = this;

			// Load
			try {
				pulverizr = require("pulverizr-bal");
			}
			catch ( e ) {
				console.log('Failed to load the [pulverizr-bal] package. Images will not be compressed.');
				//process.exit(1);
				return false;
			}

			// Compress
			config.files.img.each(function(key,filePath){
				if ( !filePath ) return; // continue
				console.log("Compressed: "+filePath);
				var
					fileOutPath = me.expandPath(filePath,config.rootPath,config.directories.out);
				pulverizr.compress(fileOutPath,{
					"quiet": true
				});
			});
		},

		/**
		 * Compress
		 */
		compress: function() {
			// Prepare
			var config = this.config;

			// JS
			if ( config.compress.js ) {
				console.log("\nCompressing JS Files:");
				this.compressJs();
			}

			// CSS
			if ( config.compress.css ) {
				console.log("\nCompressing CSS Files:");
				this.compressCss();
			}

			// IMG
			if ( config.compress.img ) {
				console.log("\nCompressing Image Files:");
				this.compressImg();
			}
		},

		/**
		 * Run the Application
		 */
		run: function() {
			// Compile
			this.compile();

			// Done
			return true;
		}

	});

	// Export
	module.exports.Buildr = Buildr;
})();
